/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Panel;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.swing.JRootPane;


import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Scale;
import org.eclipse.swt.widgets.Shell;

import core.Loader;

import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class creates a new dialog
 * with the necessary tools for the user to fit
 * a given image a specified face mask
 */
public class FitImage {

	/**
	 * Class that does all the drawing
	 */
	private class DrawingContainer extends Component {

		/**
		 * Every class that extends Component should have a serial version
		 */
		private static final long serialVersionUID = 5446285726846365378L;

		/**
		 * Holds the temporarily small image to speed up image processing
		 */
		private BufferedImage resizedImage = null;

		/**
		 * Holds the resized image with the correct rotation to speed up ptocessing
		 */
		private BufferedImage resizedRotatedImage;

		/**
		 * Backup copy of the resized image because of the scale method
		 */
		private BufferedImage resizedImageOriginal = null;

		/**
		 * Image that holds the face template
		 */
		private BufferedImage templateImage = null;

		/**
		 * The offset for the resized image that is adjusted by the user
		 */
		private java.awt.Point offset = null;

		/**
		 * Auxiliary variable to ease the image dragging
		 */
		private java.awt.Point oldOffset = null;

		/**
		 * The rotation angle that is to be adjusted by the user
		 */
		private double imageAngle = 0.0;

		//Auxiliary variables used to ease the image dragging
		private int startX;
		private int startY;

		/**
		 * The class constructor that recieve's the original image,
		 * scales-it and add all the used mouse listeners
		 * @param originalImage the image that should be fitted
		 */
		public DrawingContainer (BufferedImage originalImage) {

			//initialize the template as an empty image
			templateImage = new BufferedImage(270, 264, BufferedImage.TYPE_INT_ARGB);

			//default offset that is to be used at startup
			// TODO default offset should make image appear in the middle of the drawing area
			offset = new java.awt.Point(0,0);

			//make an backup copy of the resized image (because of scaling)
			double resizeFactor = ImageUtils.getResizingRatio(originalImage, DRAWING_AREA_WIDTH * 2, DRAWING_AREA_HEIGHT * 2);
			resizedImageOriginal = resizeImage(originalImage, resizeFactor,resizeFactor);

			//resize the original image
			resizeFactor = ImageUtils.getResizingRatio(originalImage, DRAWING_AREA_WIDTH, DRAWING_AREA_HEIGHT);
			resizedImage = resizeImage(originalImage, resizeFactor,resizeFactor);

			//rotate the image to create the final image that is to be drawn
			rotateImage();

			addMouseListener(new MouseAdapter() { //mouse listeners

				//fires when the mouse enters the drawing area, change the cursor icon
				public void mouseEntered(final MouseEvent evt) {
					setCursor (new Cursor (Cursor.HAND_CURSOR));
				}

				//fires when the mouse exit's the drawing area, restore the original cursor icon
				public void mouseExited(final MouseEvent evt) {
					setCursor (new Cursor (Cursor.DEFAULT_CURSOR));
				}

				//fires when the mouse is released, restores drawing area's default cursor icon
				public void mouseReleased(final MouseEvent evt) {
					setCursor (new Cursor (Cursor.HAND_CURSOR));
				}

				//when mouse is pressed initialize the x,y coordinates and change cursor's icon
				public void mousePressed(final MouseEvent evt) {
					startX = evt.getX();
					startY = evt.getY();
					oldOffset = new java.awt.Point(offset);
					setCursor (new Cursor (Cursor.MOVE_CURSOR));
				}
			});

			addMouseMotionListener(new MouseMotionAdapter() { //mouse motion listener

				//fires when the image is dragged in the drawing area
				public void mouseDragged(final MouseEvent evt) {

					if (setImageOffset(evt.getX() - startX + oldOffset.x, evt.getY() - startY + oldOffset.y)) {
						repaint();
					}
				}
			});
		}
		
		public void disposeImages() {
			templateImage = null;
			resizedImage = null;
			resizedImageOriginal = null;
			resizedRotatedImage = null;
		}

		/**
		 * Change the resizedImage offset on screen to the indicated one
		 * Values out of range are filtered to borders 
		 * @param x new x offset value
		 * @param y new y offset value
		 * @return true if the original image offset was changed
		 */
		private boolean setImageOffset(int x, int y) {
			
			//whether we made changes to the original offset of the image
			boolean offsetHasChanged = false;
			
			//the number of pixels that we keep to maintain a portion of the image always
			//visible in the drawing area
			int minVisiblePixels = 40;

			//only allow the image to be dragged if is not outside the drawing area
			if (x < DRAWING_AREA_WIDTH - minVisiblePixels && -x < resizedImage.getWidth() - minVisiblePixels) {
				if (offset.x != x) {
					offset.x =  x;
					offsetHasChanged = true;
				}
			} else { //if we try to drag the image too far from the drawing area use only the max allowed
				if (x >= DRAWING_AREA_WIDTH - minVisiblePixels) {
					if (offset.x != DRAWING_AREA_WIDTH - minVisiblePixels) {
						offset.x =  DRAWING_AREA_WIDTH - minVisiblePixels;
						offsetHasChanged = true;
					}
				} else {
					if (offset.x != -(resizedImage.getWidth() - minVisiblePixels)) {
						offset.x =  -(resizedImage.getWidth() - minVisiblePixels);
						offsetHasChanged = true;
					}
				}
			}

			//only allow the image to be dragged if is not outside the drawing area
			if (y < DRAWING_AREA_HEIGHT - minVisiblePixels && -y < resizedImage.getHeight() - minVisiblePixels) {
				if (offset.y != y) {
					offset.y =  y;
					offsetHasChanged = true;
				}
			} else { //if we try to drag the image too far from the drawing area use only the max allowed
				if (y >= DRAWING_AREA_HEIGHT - minVisiblePixels) {
					if (offset.y != DRAWING_AREA_HEIGHT - minVisiblePixels) {
						offset.y =  DRAWING_AREA_HEIGHT - minVisiblePixels;
						offsetHasChanged = true;
					}
				} else {
					if (offset.y != -(resizedImage.getHeight() - minVisiblePixels)) {
						offset.y =  -(resizedImage.getHeight() - minVisiblePixels);
						offsetHasChanged = true;
					}
				}
			}

			return offsetHasChanged;
		}

		/**
		 * Method that resize an given image using the width and height factor
		 * @param image the image that is to be resized
		 * @param widthFactor the width factor
		 * @param heightFactor the height factor
		 * @return the resized image
		 */
		private BufferedImage resizeImage(BufferedImage image, double widthFactor, double heightFactor) {
			AffineTransform tx = new AffineTransform(); //use affine transform class to do the resizing
			tx.scale(widthFactor, heightFactor);
			AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);
			return op.filter(image, null);
		}

		/**
		 * Method that draws the image behind the face template on an white background
		 */
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;

			//draw the white background
			g2d.setColor(Color.white);
			g2d.fillRect(0,0, getWidth(), getHeight());

			//use antialias for drawings
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
			g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);

			//draw the final result
			g.drawImage(resizedRotatedImage, offset.x,offset.y, null);
			g.drawImage(templateImage, (DRAWING_AREA_WIDTH / 2) - (templateImage.getWidth() / 2), (DRAWING_AREA_HEIGHT / 2) - (templateImage.getHeight() / 2), null);
		}

		private void rotateImage() {
			AffineTransform tx = new AffineTransform();
			tx.rotate(Math.toRadians(imageAngle), resizedImage.getWidth() / 2.0, resizedImage.getHeight() / 2.0);
			AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);
			resizedRotatedImage = op.filter(resizedImage, null);
		}

		/**
		 * Method that initializes the templateImage with the given template
		 * @param name
		 */
		public void setTemplate(String name) {
			try { //read the template from the resources folder
				templateImage = ImageIO.read(new File(System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + name + ".png"));
				repaint(); //update the drawing area with the new template
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		/**
		 * Move the image one pixel to the left
		 */
		public void moveLeft() {
			if (setImageOffset(offset.x - 1, offset.y)) {
				repaint();
			}
		}

		/**
		 * Move the image one pixel to the right
		 */
		public void moveRight() {
			if (setImageOffset(offset.x + 1, offset.y)) {
				repaint();
			}
		}

		/**
		 * Move the image up with one pixel
		 */
		public void moveUp() {
			if (setImageOffset(offset.x, offset.y - 1)) {
				repaint();
			}
		}

		/**
		 * Move the image down with one pixel
		 */
		public void moveDown() {
			if (setImageOffset(offset.x, offset.y + 1)) {
				repaint();
			}
		}

		/**
		 * Method that adjust's the image scale
		 * @param scaleValue the value
		 */
		public void adjustScale(int scaleValue) {
			scaleValue -= 15;
			double widthFactor = ((double)scaleValue / 35) + 1;
			double heightFactor = ((double)scaleValue / 35) + 1;
			if (widthFactor > 0) {
				resizedImage = resizeImage(resizedImageOriginal, widthFactor / 2, heightFactor / 2);
				rotateImage();
				setImageOffset(offset.x, offset.y);
				repaint();
			}
		}

		/**
		 * Method that adjust's the image rotation
		 * @param angleValue the value (min: 0 = -180°, max: 180 = +180°)
		 */
		public void adjustAngle(int angleValue) {
			//multiply 2 and subtract 180° from the value because slider has a range of 0° - 180°
			angleValue = (angleValue * 2) - 180;
			this.imageAngle = angleValue;
			rotateImage();
			repaint(); //make changes visible
		}
	}

	//the width and height for the drawing area class
	private static final int DRAWING_AREA_WIDTH = 240;
	private static final int DRAWING_AREA_HEIGHT = 320;

	/**
	 * The folder inside the pixmaps folder where the face templates are located
	 */
	private static final String FACE_TEMPLATES_FOLDER = "face_templates";

	/**
	 * The shell for the dialog window
	 */
	private Shell sShell = null;

	/**
	 * The composite to hold the {@link SWT#EMBEDDED} composite
	 */
	private Composite compositeImageArea = null;

	/**
	 * The composite for all the dialog controls
	 * (template buttons, scale, fine adjustment)
	 */
	private Composite compositeControls = null;

	/**
	 * The bottom ok/cancel composite
	 */
	private Composite compositeOkCancel = null;

	/**
	 * The composite to hold the SWING/AWT component
	 */
	private Composite compositeAWT_SWT = null;

	/**
	 * Label for the face template buttons
	 */
	private Label labelFaceTemplates = null;

	/**
	 * Composite to display the image title label
	 */
	private Composite compositeImageTitle = null;

	/**
	 * Composite used just to give a top indentation for the window
	 */
	private Composite compositeTopIdentation = null;

	//ok/cancel buttons
	private Button buttonCancel = null;
	private Button buttonOk = null;

	/**
	 * Composite to hold the buttons with the face templates
	 */
	private Composite compositeFaceTemplates = null;

	//face templates buttons
	private Button buttonFaceTemplate1 = null;
	private Button buttonFaceTemplate2 = null;
	private Button buttonFaceTemplate3 = null;
	private Button buttonFaceTemplate4 = null;

	/**
	 * Composite for the size and rotation scale
	 */
	private Composite compositeSizeRotation = null;

	//size and rotation scale
	private Scale scaleSize = null;
	private Scale scaleRotation = null;

	//labels for the size and rotation scale
	private Label labelSize = null;
	private Label labelRotation = null;

	/**
	 * Composite that holds the image example and the fine adjust composite
	 */
	private Composite compositeBottomControls = null;

	//labels for compositeBottomControls
	private Label labelExample = null;
	private Label labelFineAdjust = null;

	/**
	 * Composite that holds the adjust buttons
	 */
	private Composite compositeFineAdjustButtons = null;

	//fine adjust buttons
	private Button buttonMoveLeft = null;
	private Button buttonMoveUp = null;
	private Button buttonMoveRight = null;
	private Button buttonMoveDown = null;

	/**
	 * Canvas for the example image
	 */
	private Canvas canvasExampleImage = null;

	/**
	 * The SWING/AWT drawing container
	 */
	private DrawingContainer dc;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * key listener used to notify the shell that a key was pressed
	 */
	private Listener sShellKeyListener;

	/**
	 * Composite used to display the quick help about this dialog
	 */
	private Composite compositeTips;

	//the title and text for the tips
	private Label labelTitle;
	private Label labelText;


	//TODO help button and tips for taking the pictures with the camera
	/**
	 * Creates a new instance of this class given its parent
	 * and the desired image to be adjusted
	 */
	public FitImage(Shell parent, BufferedImage originalImage) {
		dc = new DrawingContainer(originalImage);
		messages = Loader.getMessages();
		createSShell(parent);
	}

	/**
	 * This method initializes sShell
	 * @param originalImage 
	 * @param parent 
	 */
	private void createSShell(Shell parent) {

		//shell grid layout
		GridLayout gridLayoutSShell = new GridLayout();
		gridLayoutSShell.numColumns = 2;
		gridLayoutSShell.verticalSpacing = 0;
		gridLayoutSShell.horizontalSpacing = 0;

		//initialize the shell
		sShell = new Shell(parent, SWT.APPLICATION_MODAL | SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("PLUGIN-FIT_IMAGE_TITLE"));
		sShell.setLayout(gridLayoutSShell);

		//sometimes the sShellKeyListener is not called because sShell has no focus but is active. this will fix it
		sShell.addShellListener(new ShellAdapter() {
			public void shellActivated(ShellEvent e) {
				sShell.setFocus();
				super.shellActivated(e);
			}
		});

		//listener used to notify the shell that ESC key was pressed
		sShellKeyListener = new Listener() {
			public void handleEvent(Event event) {
				//certify that sShell is the active shell and the ESC was pressed
				if (sShell.getDisplay().getActiveShell().equals(sShell) && event.keyCode == SWT.ESC) { 
					sShell.close();
				}
			}
		};
		sShell.getDisplay().addFilter(SWT.KeyDown, sShellKeyListener);

		//remove the listener when shell is disposed
		sShell.addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				sShell.getDisplay().removeFilter(SWT.KeyDown, sShellKeyListener);
			}
		});

		//create the shell components and pack the shell
		createCompositeTopIdentation();
		createCompositeImageTitle();
		createCompositeImageArea();
		createCompositeControls();
		createCompositeTips();
		createCompositeOkCancel();
		sShell.pack();
	}

	/**
	 * This method initializes compositeImageArea	
	 */
	private void createCompositeImageArea() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeImageArea = new Composite(sShell, SWT.NONE);
		compositeImageArea.setLayout(new GridLayout());
		compositeImageArea.setLayoutData(gridDataComposite);

		//create the composite's children
		createCompositeAWT_SWT();
	}

	/**
	 * This method initializes compositeControls
	 */
	private void createCompositeControls() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 1;
		gridLayoutComposite.marginWidth = 0;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.horizontalIndent = 5;
		gridDataComposite.grabExcessVerticalSpace = true;

		//grid data for the face templates label
		GridData gridDataLabelFaceTemplates = new GridData();
		gridDataLabelFaceTemplates.grabExcessHorizontalSpace = true;
		gridDataLabelFaceTemplates.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeControls = new Composite(sShell, SWT.NONE);
		compositeControls.setLayoutData(gridDataComposite);
		compositeControls.setLayout(gridLayoutComposite);

		//initialize the face templates label
		labelFaceTemplates = new Label(compositeControls, SWT.NONE);
		labelFaceTemplates.setText(messages.getString("PLUGIN-FIT_IMAGE_CHOOSE_TEMPLATE"));
		labelFaceTemplates.setLayoutData(gridDataLabelFaceTemplates);

		//create the composite's children
		createCompositeFaceTemplates();
		createCompositeSizeRotation();
		createCompositeBottomControls();
	}

	/**
	 * This method initializes compositeTips	
	 */
	private void createCompositeTips() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.grabExcessVerticalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.horizontalIndent = 5;
		gridDataComposite.horizontalSpan = 2;

		//grid data for the label title
		GridData gridDataLabelTitle = new GridData();
		gridDataLabelTitle.horizontalIndent = 5;
		gridDataLabelTitle.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataLabelTitle.grabExcessVerticalSpace = true;

		//grid data for the label text
		GridData gridDataLabelText = new GridData();
		gridDataLabelText.verticalAlignment = org.eclipse.swt.layout.GridData.BEGINNING;
		gridDataLabelText.grabExcessVerticalSpace = true;
		gridDataLabelText.horizontalIndent = 5;

		//initialize the composite
		compositeTips = new Composite(sShell, SWT.BORDER);
		compositeTips.setLayout(new GridLayout());
		compositeTips.setLayoutData(gridDataComposite);

		//the title and the text labels
		labelTitle = new Label(compositeTips, SWT.NONE);
		labelTitle.setText(messages.getString("PLUGIN-FIT_IMAGE_TIP_TITLE"));
		labelTitle.setLayoutData(gridDataLabelTitle);
		labelTitle.setFont(new Font(Display.getDefault(), "Sans", 14, SWT.BOLD));
		labelText = new Label(compositeTips, SWT.NONE);
		labelText.setText(messages.getString("PLUGIN-FIT_IMAGE_TIP_TEXT"));
		labelText.setLayoutData(gridDataLabelText);
	}

	/**
	 * This method initializes compositeOkCancel
	 */
	private void createCompositeOkCancel() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;
		gridLayoutComposite.marginWidth = 1;
		gridLayoutComposite.marginHeight = 1;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalSpan = 2;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.verticalIndent = 5;

		//grid data for the ok button
		GridData gridDataOk = new GridData();
		gridDataOk.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataOk.grabExcessVerticalSpace = true;

		//grid data for the cancel button
		GridData gridDataCancel = new GridData();
		gridDataCancel.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataCancel.grabExcessVerticalSpace = true;

		//initialize the composite
		compositeOkCancel = new Composite(sShell, SWT.NONE);
		compositeOkCancel.setLayoutData(gridDataComposite);
		compositeOkCancel.setLayout(gridLayoutComposite);

		//initialize the ok/cancel buttons
		buttonCancel = new Button(compositeOkCancel, SWT.NONE);
		buttonCancel.setText(messages.getString("PLUGIN-FIT_IMAGE_CANCEL"));
		buttonCancel.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CANCEL, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonCancel.setLayoutData(gridDataCancel);
		buttonOk = new Button(compositeOkCancel, SWT.NONE);
		buttonOk.setText(messages.getString("PLUGIN-FIT_IMAGE_OK"));
		buttonOk.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_OK, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonOk.setLayoutData(gridDataOk);

		//adjust the width of the buttons to be the same on all of them
		gridDataOk.widthHint = Math.max(buttonCancel.computeSize(-1, -1).x, buttonOk.computeSize(-1, -1).x);
		gridDataCancel.widthHint = Math.max(buttonCancel.computeSize(-1, -1).x, buttonOk.computeSize(-1, -1).x);

		//listeners for the two buttons
		buttonCancel.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				sShell.close();
			}
		});
		buttonOk.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				sShell.dispose();
			}
		});
	}

	/**
	 * This method initializes compositeAET_SWT
	 */
	private void createCompositeAWT_SWT() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.widthHint = DRAWING_AREA_WIDTH;
		gridDataComposite.heightHint = DRAWING_AREA_HEIGHT;

		//initialize the composite to hold the SWING/AWT frame
		compositeAWT_SWT = new Composite(compositeImageArea, SWT.BORDER | SWT.EMBEDDED);
		compositeAWT_SWT.setLayout(new GridLayout());
		compositeAWT_SWT.setLayoutData(gridDataComposite);

		//create the SWING/AWT frame and containers
		Frame frame = SWT_AWT.new_Frame(compositeAWT_SWT);
		frame.setBackground(Color.white);
		Panel panel = new Panel(new BorderLayout());
		frame.add(panel);
		JRootPane root = new JRootPane();
		panel.add(root);
		java.awt.Container contentPane = root.getContentPane();
		contentPane.setLayout(new BorderLayout());
		contentPane.add(dc, BorderLayout.CENTER);
	}

	/**
	 * This method initializes compositeImageTitle
	 */
	private void createCompositeImageTitle() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.marginHeight = 0;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalSpan = 2;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessHorizontalSpace = true;

		//grid data for the label
		GridData gridDataLabel = new GridData();
		gridDataLabel.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataLabel.grabExcessHorizontalSpace = true;

		//initialize the composite
		compositeImageTitle = new Composite(sShell, SWT.NONE);
		compositeImageTitle.setLayoutData(gridDataComposite);
		compositeImageTitle.setLayout(gridLayoutComposite);

		//initialize the label
		Label label = new Label(compositeImageTitle, SWT.NONE);
		label.setText(messages.getString("PLUGIN-FIT_IMAGE_IMAGE_PREVIEW"));
		label.setLayoutData(gridDataLabel);
	}

	/**
	 * This method initializes compositeTopIdentation
	 */
	private void createCompositeTopIdentation() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.horizontalSpan = 2;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeTopIdentation = new Composite(sShell, SWT.NONE);
		compositeTopIdentation.setLayout(new GridLayout());
		compositeTopIdentation.setLayoutData(gridDataComposite);
	}

	/**
	 * This method initializes compositeFaceTemplates
	 */
	private void createCompositeFaceTemplates() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 4;
		gridLayoutComposite.marginWidth = 0;
		gridLayoutComposite.marginHeight = 0;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeFaceTemplates = new Composite(compositeControls, SWT.NONE);
		compositeFaceTemplates.setLayoutData(gridDataComposite);
		compositeFaceTemplates.setLayout(gridLayoutComposite);

		//initialize the face template buttons
		buttonFaceTemplate1 = new Button(compositeFaceTemplates, SWT.NONE);
		buttonFaceTemplate1.setImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + "heart-small.jpg"));
		buttonFaceTemplate1.setToolTipText(messages.getString("PLUGIN-FIT_IMAGE_HEART_TEMPLATE"));
		buttonFaceTemplate2 = new Button(compositeFaceTemplates, SWT.NONE);
		buttonFaceTemplate2.setImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + "oval-small.jpg"));
		buttonFaceTemplate2.setToolTipText(messages.getString("PLUGIN-FIT_IMAGE_OVAL_TEMPLATE"));
		buttonFaceTemplate3 = new Button(compositeFaceTemplates, SWT.NONE);
		buttonFaceTemplate3.setImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + "round-small.jpg"));
		buttonFaceTemplate3.setToolTipText(messages.getString("PLUGIN-FIT_IMAGE_ROUND_TEMPLATE"));
		buttonFaceTemplate4 = new Button(compositeFaceTemplates, SWT.NONE);
		buttonFaceTemplate4.setImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + "square-small.jpg"));
		buttonFaceTemplate4.setToolTipText(messages.getString("PLUGIN-FIT_IMAGE_SQUARE_TEMPLATE"));

		//listeners for the face template buttons
		buttonFaceTemplate1.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.setTemplate("heart");
			}
		});
		buttonFaceTemplate2.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.setTemplate("oval");
			}
		});
		buttonFaceTemplate3.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.setTemplate("round");
			}
		});
		buttonFaceTemplate4.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.setTemplate("square");
			}
		});
	}

	/**
	 * This method initializes compositeSizeRotation
	 */
	private void createCompositeSizeRotation() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.marginWidth = 0;
		gridLayoutComposite.numColumns = 2;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.grabExcessVerticalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the rotation scale
		GridData gridDataScaleRotation = new GridData();
		gridDataScaleRotation.grabExcessHorizontalSpace = true;
		gridDataScaleRotation.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the size scale
		GridData gridDataScaleSize = new GridData();
		gridDataScaleSize.grabExcessHorizontalSpace = true;
		gridDataScaleSize.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeSizeRotation = new Composite(compositeControls, SWT.NONE);
		compositeSizeRotation.setLayoutData(gridDataComposite);
		compositeSizeRotation.setLayout(gridLayoutComposite);

		//initialize the size label and scale
		labelSize = new Label(compositeSizeRotation, SWT.NONE);
		labelSize.setText(messages.getString("PLUGIN-FIT_IMAGE_SIZE"));
		scaleSize = new Scale(compositeSizeRotation, SWT.NONE);
		scaleSize.setLayoutData(gridDataScaleSize);
		scaleSize.setMinimum(0);
		scaleSize.setSelection(15);
		scaleSize.setMaximum(100);

		//initialize the rotation label and scale
		labelRotation = new Label(compositeSizeRotation, SWT.NONE);
		labelRotation.setText(messages.getString("PLUGIN-FIT_IMAGE_ROTATION"));
		scaleRotation = new Scale(compositeSizeRotation, SWT.NONE);
		scaleRotation.setLayoutData(gridDataScaleRotation);
		scaleRotation.setMinimum(0);
		scaleRotation.setSelection(90);
		scaleRotation.setMaximum(180);

		//scale listeners
		scaleSize.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				dc.adjustScale(scaleSize.getSelection());
			}
		});
		scaleRotation.addListener(SWT.Selection, new Listener() {
			public void handleEvent(Event event) {
				dc.adjustAngle(scaleRotation.getSelection());
			}
		});
	}

	/**
	 * This method initializes compositeBottomControls
	 */
	private void createCompositeBottomControls() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.marginWidth = 0;
		gridLayoutComposite.marginHeight = 0;
		gridLayoutComposite.numColumns = 2;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the composite
		compositeBottomControls = new Composite(compositeControls, SWT.NONE);
		compositeBottomControls.setLayoutData(gridDataComposite);
		compositeBottomControls.setLayout(gridLayoutComposite);

		//initialize the labels
		labelExample = new Label(compositeBottomControls, SWT.NONE);
		labelExample.setText(messages.getString("PLUGIN-FIT_IMAGE_EXAMPLE"));
		labelFineAdjust = new Label(compositeBottomControls, SWT.NONE);
		labelFineAdjust.setText(messages.getString("PLUGIN-FIT_IMAGE_FINE_ADJUSTMENT"));

		//the composite's children
		createCanvasExampleImage();
		createCompositeFineAdjust();
	}

	/**
	 * This method initializes compositeFineAdjust
	 */
	private void createCompositeFineAdjust() {

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 3;
		gridLayoutComposite.horizontalSpacing = 0;
		gridLayoutComposite.verticalSpacing = 0;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.END;

		//initialize the composite
		compositeFineAdjustButtons = new Composite(compositeBottomControls, SWT.BORDER);
		compositeFineAdjustButtons.setLayout(gridLayoutComposite);
		compositeFineAdjustButtons.setLayoutData(gridDataComposite);

		//initialize the move up button
		@SuppressWarnings("unused")
		Label filler = new Label(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveUp = new Button(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveUp.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_UP, StockIcon.SIZE_SMALL_TOOLBAR, true)));

		//initialize the move left button
		@SuppressWarnings("unused")
		Label filler2 = new Label(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveLeft = new Button(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveLeft.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_BACK, StockIcon.SIZE_SMALL_TOOLBAR, true)));

		//initialize the move right button
		@SuppressWarnings("unused")
		Label filler1 = new Label(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveRight = new Button(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveRight.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_FORWARD, StockIcon.SIZE_SMALL_TOOLBAR, true)));

		//initialize the move down button
		@SuppressWarnings("unused")
		Label filler3 = new Label(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveDown = new Button(compositeFineAdjustButtons, SWT.NONE);
		buttonMoveDown.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_DOWN, StockIcon.SIZE_SMALL_TOOLBAR, true)));

		//buttons listeners
		buttonMoveUp.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.moveUp();
			}
		});
		buttonMoveLeft.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.moveLeft();
			}
		});
		buttonMoveRight.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.moveRight();
			}
		});
		buttonMoveDown.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				dc.moveDown();
			}
		});
	}

	/**
	 * This method initializes canvasExampleImage
	 */
	private void createCanvasExampleImage() {

		//grid data for the canvas
		GridData gridDataCanvas = new GridData();
		gridDataCanvas.verticalAlignment = org.eclipse.swt.layout.GridData.BEGINNING;
		gridDataCanvas.grabExcessHorizontalSpace = true;
		gridDataCanvas.horizontalAlignment = org.eclipse.swt.layout.GridData.BEGINNING;

		//initialize the canvas
		canvasExampleImage = new Canvas(compositeBottomControls, SWT.BORDER);
		canvasExampleImage.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + FACE_TEMPLATES_FOLDER + File.separator + "perfect_fit.png"));
		canvasExampleImage.setLayoutData(gridDataCanvas);

		//adjust the canvas size to the given image size
		//TODO -2 values are because SWT.BORDER adds those pixels in composite. why!?
		gridDataCanvas.widthHint = canvasExampleImage.getBackgroundImage().getBounds().width - 2;
		gridDataCanvas.heightHint = canvasExampleImage.getBackgroundImage().getBounds().height - 2;
	}

	/**
	 * Makes the dialog visible and brings it to the front
	 * of the display.
	 *
	 * @return an array of a Rectangle with the image area,
	 *         a Dimension with image size and
	 *         a double specifying the image rotation angle
	 *         or <code>null</code>if the dialog was canceled
	 *         or an error occurred
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the dialog has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the dialog</li>
	 * </ul>
	 */
	public Object[] open() {

		//open the shell and wait to finish
		sShell.open();
		sShell.addShellListener(new org.eclipse.swt.events.ShellAdapter() {
			public void shellClosed(org.eclipse.swt.events.ShellEvent e) {
				dc.offset = null;
			}
		});

		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}
		if (dc.offset == null) {
			dc.disposeImages();
			return null;
		} else {
			Rectangle imageArea = new Rectangle(-dc.offset.x, -dc.offset.y, DRAWING_AREA_WIDTH, DRAWING_AREA_HEIGHT);
			Dimension imageSize = new Dimension(dc.resizedImage.getWidth(), dc.resizedImage.getHeight());
			dc.disposeImages();
			return new Object[] {imageArea, imageSize, dc.imageAngle};
		}
	}
}