/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.ResourceBundle;


import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Scale;

import core.Loader;

import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class represents a tab that
 * can be used to paint the face lips
 *
 */
public class LipsTab extends TabContainer {

	/**
	 * Instances of this class represents a state of the {@link LipsTab}
	 */
	private class LipsHistoryStep {

		/**
		 * The selected button
		 */
		private final ColorButton selectedColorButton;

		/**
		 * The opacity value used to paint the lips
		 */
		private final int lipsOpacityScaleValue;

		/**
		 * Creates a new instance of this class to save the
		 * {@link LipsTab} given state
		 * @param colorButton the selected button
		 * @param lipsOpacityScaleValue the opacity value used to paint the lips
		 */
		public LipsHistoryStep(ColorButton colorButton, int lipsOpacityScaleValue) {

			//search for the reference of the actual button instead of the reference of the selectedButton
			ColorButton selectedColorButton = null;
			if (colorButton != null) {
				for (Control c : colorButton.getParent().getChildren()) {
					if (c instanceof ColorButton) {
						ColorButton cb = (ColorButton) c;
						if (cb.equals(colorButton)) {
							selectedColorButton = cb;
							break;
						}
					}
				}
			}
			this.selectedColorButton = selectedColorButton;
			this.lipsOpacityScaleValue = lipsOpacityScaleValue;
		}
	}

	/**
	 * The opacity that will be used to draw the lips color
	 */
	private static float ALPHA_COMPOSITE_OPACITY = 0.2f;

	/**
	 * Alpha composite used to draw the lips image
	 */
	private AlphaComposite currentCompositeOpacity;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * A list with all the colors used by ColorButton
	 * For each color in this list one ColorButton will be created
	 */
	private Color[] colorsList = {
			new Color(Display.getCurrent(), 255, 212, 190),
			new Color(Display.getCurrent(), 253, 200, 146),
			new Color(Display.getCurrent(), 242, 164, 115),
			new Color(Display.getCurrent(), 238, 137, 109),
			new Color(Display.getCurrent(), 235, 123, 62),
			new Color(Display.getCurrent(), 183, 108, 69),
			new Color(Display.getCurrent(), 255, 151, 214),
			new Color(Display.getCurrent(), 246, 132, 158),
			new Color(Display.getCurrent(), 234, 116, 143),
			new Color(Display.getCurrent(), 241, 102, 185),
			new Color(Display.getCurrent(), 245, 53, 167),
			new Color(Display.getCurrent(), 187, 83, 134),
			new Color(Display.getCurrent(), 255, 183, 183),
			new Color(Display.getCurrent(), 255, 158, 158),
			new Color(Display.getCurrent(), 255, 136, 136),
			new Color(Display.getCurrent(), 255, 95, 96),
			new Color(Display.getCurrent(), 239, 92, 125),
			new Color(Display.getCurrent(), 244, 64, 106),
			new Color(Display.getCurrent(), 255, 16, 16),
			new Color(Display.getCurrent(), 234, 13, 13),
			new Color(Display.getCurrent(), 215, 8, 8),
			new Color(Display.getCurrent(), 188, 0, 0),
			new Color(Display.getCurrent(), 154, 0, 0),
			new Color(Display.getCurrent(), 146, 40, 65),
			new Color(Display.getCurrent(), 164, 162, 177),
			new Color(Display.getCurrent(), 154, 121, 111),
			new Color(Display.getCurrent(), 142, 94, 80),
			new Color(Display.getCurrent(), 135, 68, 47),
			new Color(Display.getCurrent(), 103, 47, 30),
			new Color(Display.getCurrent(), 0, 0, 0)
	};

	/**
	 * Used to set the upper and color buttons composite
	 */
	private TabGuiUtility tabGui;

	/**
	 * the tab's main composite
	 */
	private Composite mainComposite;

	/**
	 * All the colored buttons in use by this tab
	 */
	private ArrayList<ColorButton> coloredButtonList;

	//mouse listeners in the drawing area
	private MouseListener mouseListener;
	private MouseMotionListener mouseMotionListener;

	/**
	 * Composite for the bottom controls
	 */
	private Composite bottomComposite;

	//bottom composite label separators
	private Label labelSeparatorUp;
	private Label labelSeparatorDown;

	/**
	 * Scale for the opacity of the lips image
	 */
	private Scale opacityScale;

	/**
	 * Label for the opacity scale
	 */
	private Label labelOpacityScale;

	/**
	 * Button for clearing this tab's contents on the drawing area
	 */
	private Button clearButton;

	/**
	 * The tab's final lips image
	 */
	private BufferedImage lipsImage;

	/**
	 * flag to indicate whether the tab has or not access to call for redrawing the image area
	 */
	protected boolean imageAreaRedrawAllowed;

	/**
	 * A reference to this instance class
	 */
	private LipsTab thisLipsTab = this;

	/**
	 * Path that contains the lips area
	 */
	private GeneralPath lipsGeneralPath;

	/**
	 * Constructs a new instance of this class given its parent,
	 * a style value describing its behavior and appearance and
	 * this instance's position in the {@link TabsContainer} array
	 *
	 * @param parent a {@link TabsContainer} which will be the parent of the new instance
	 * @param style the style of the tab
	 * @param tabLayerPosition the tab's position in the {@link TabsContainer} array
	 *
	 */
	public LipsTab(TabsContainer parent, int style, int tabLayerPosition) {
		super(parent, style, tabLayerPosition);

		messages = Loader.getMessages();

		//the 2 things needed to be done by every implementation of TabComposite
		super.setTabImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + "tab_icon_lips.png"));
		super.setTabToolTipText(messages.getString("LIPS_TAB_TOOLTIP"));

		addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				for (Color c : colorsList) {
					c.dispose();
				}
			}
		});

		imageAreaRedrawAllowed = true;

		//create the tab's GUI
		tabGui = new TabGuiUtility(this);
		mainComposite = this;
		tabGui.createUpperComposite(super.getTabToolTipText());
		coloredButtonList = tabGui.createColorButtonsComposite(colorsList);
		tabGui.setSelectedColorButton(null);
		createBottomComposite();

		//default opacity
		currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ALPHA_COMPOSITE_OPACITY);

		extendColorButtonsBehavior();
	}

	/**
	 * This method initializes bottomComposite	
	 */
	public void createBottomComposite() {

		//grid data for the clear button
		GridData gridDataClearButton = new GridData();
		gridDataClearButton.grabExcessHorizontalSpace = true;
		gridDataClearButton.horizontalSpan = 2;
		gridDataClearButton.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		
		//grid data for the separator label up
		GridData gridDataSeparatorLabelUp = new GridData();
		gridDataSeparatorLabelUp.grabExcessHorizontalSpace = true;
		gridDataSeparatorLabelUp.horizontalSpan = 2;
		gridDataSeparatorLabelUp.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label down
		GridData gridDataSeparatorLabelDown = new GridData();
		gridDataSeparatorLabelDown.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataSeparatorLabelDown.horizontalSpan = 2;
		gridDataSeparatorLabelDown.grabExcessHorizontalSpace = true;
		
		//grid data for the opacity scale
		GridData gridDataOpacityScale = new GridData();
		gridDataOpacityScale.grabExcessHorizontalSpace = true;
		gridDataOpacityScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		
		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;
		
		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;
		
		//initialize the composite
		bottomComposite = new Composite(mainComposite, SWT.NONE);
		bottomComposite.setLayout(gridLayoutComposite);
		bottomComposite.setLayoutData(gridDataComposite);
		
		//initialize the upper label separator
		labelSeparatorUp = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorUp.setLayoutData(gridDataSeparatorLabelUp);
		
		//initialize the clear button
		clearButton = new Button(bottomComposite, SWT.NONE);
		clearButton.setText(messages.getString("LIPS_TAB_CLEAR"));
		clearButton.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLEAR, StockIcon.SIZE_MENU, true)));
		clearButton.setLayoutData(gridDataClearButton);
		
		//initialize the opacity scale label
		labelOpacityScale = new Label(bottomComposite, SWT.NONE);
		labelOpacityScale.setText(messages.getString("LIPS_TAB_OPACITY"));
		
		//initialize the opacity scale
		opacityScale = new Scale(bottomComposite, SWT.NONE);
		opacityScale.setSelection(50);
		opacityScale.setLayoutData(gridDataOpacityScale);
		
		//initialize the lower separator label
		labelSeparatorDown = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorDown.setLayoutData(gridDataSeparatorLabelDown);

		//clear button listener
		clearButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//TODO check if history step is created only when needed
				if (e.widget instanceof Button && lipsGeneralPath != null && tabGui.getSelectedButton() != null) {
					tabsContainer.createHistoryStep();
					clearImageArea();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisLipsTab);
					}
				}
			}
		});

		//opacity scale listeners
		opacityScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				double opacityInterval = 0.3;
				double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
				currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisLipsTab);
				}
			}
		});
		opacityScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});
	}

	/**
	 * Color buttons should do more than just updating the {@link ColorDifference}
	 * All color buttons listeners are implemented here
	 */
	private void extendColorButtonsBehavior() {
		for (final ColorButton button : coloredButtonList) {
			button.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {

				public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
					//TODO history is created but user may click cancel on LipsTracer and nothing is drawn
					tabsContainer.createHistoryStep();
				}

				public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
					if (!hasLipsPath()) {
						tabsContainer.setLipsPath();
					}
					createLipsImage();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisLipsTab);
					}
				}
			});
		}
	}

	/**
	 * Clears this tab's image in the drawing area
	 * and resets the needed variables
	 */
	private void clearImageArea() {
		lipsImage = null;
		tabGui.setSelectedColorButton(null);
	}

	/**
	 * Create this tab's lips image  by
	 * drawing the lips path into a new
	 * image using the selected button's color
	 */
	private void createLipsImage() {
		lipsImage = null;
		if (lipsGeneralPath != null && tabGui.getSelectedButton() != null) {
			lipsImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_RGB);
			Graphics2D g2d = lipsImage.createGraphics();
			g2d.setColor(tabGui.getCurrentColor());
			g2d.fillRect(0, 0, lipsImage.getWidth(), lipsImage.getHeight());
			g2d.dispose();
			lipsImage = ImageUtils.createSoftClipping(lipsImage, lipsGeneralPath);
		}
	}

	/**
	 * Set the lips path to the given one; if the new
	 * lips path is <code>null</code> nothing is done
	 * @param lipsPath the new lips path or <code>null</code>
	 */
	public void setLipsPath(GeneralPath lipsPath) {
		if (lipsPath == null) return ;
		lipsGeneralPath = lipsPath;
		createLipsImage();
		if (imageAreaRedrawAllowed) {
			tabsContainer.redrawImageArea(thisLipsTab);
		}
	}

	/**
	 * Check whether this tab instance has an associated lips path
	 * @return true if a lips path is found, false otherwise
	 */
	public boolean hasLipsPath() {
		return lipsGeneralPath != null;
	}

	@Override
	public BufferedImage getLayerImage() {
		//create a copy of this tab's image and return it
		BufferedImage transparentLipsImage = null;
		if (lipsImage != null) {
			transparentLipsImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2d = transparentLipsImage.createGraphics();
			drawLipsImage(g2d);
			g2d.dispose();
		}
		return transparentLipsImage;
	}

	/**
	 * Draws the lips image using the given graphics context
	 * @param g2d the graphics context used for drawing operations
	 */
	private void drawLipsImage(Graphics2D g2d) {

		//save the current composite
		java.awt.Composite defaultComposite = g2d.getComposite();

		//set lips image opacity
		g2d.setComposite(currentCompositeOpacity);

		//draw the lips image
		g2d.drawImage(lipsImage, 0, 0, null);

		//restore the g2d original composite
		g2d.setComposite(defaultComposite);

	}

	@Override
	public void drawCurrentLayer(Graphics2D g2d) {
		drawLipsImage(g2d);
	}

	@Override
	public void clear() {
		clearImageArea();
	}

	@Override
	public Object getCurrentState() {
		return new LipsHistoryStep(tabGui.getSelectedButton(),opacityScale.getSelection());
	}

	@Override
	public void setCurrentState(Object o) {
		if (o instanceof LipsHistoryStep) {
			LipsHistoryStep hs = (LipsHistoryStep) o;

			imageAreaRedrawAllowed = false;

			tabGui.setSelectedColorButton(hs.selectedColorButton);

			opacityScale.setSelection(hs.lipsOpacityScaleValue);
//			opacityScale.notifyListeners(eventType, event) //TODO this is how should work instead of the 3 lines below 
			double opacityInterval = 0.3;
			double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
			currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));

			createLipsImage();

			imageAreaRedrawAllowed = true;
		}
	}


	public MouseListener mouseListener() {
		if (mouseListener == null) {
			mouseListener = new MouseListener() {

				//unimplemented methods
				public void mouseClicked(MouseEvent e) {}
				public void mouseExited(MouseEvent e) {}
				public void mousePressed(MouseEvent e) {}
				public void mouseReleased(MouseEvent e) {}

				public void mouseEntered(MouseEvent e) {
					e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));					
				}
			};
		}
		return mouseListener;
	}

	public MouseMotionListener mouseMotionListener() {
		if (mouseMotionListener == null) {
			mouseMotionListener = new MouseMotionListener() {
				
				//unimplemented methods
				public void mouseDragged(MouseEvent e) {}
				public void mouseMoved(MouseEvent e) {}
			};
		}
		return mouseMotionListener;
	}

	public void originalImageChanged() {
		lipsGeneralPath = null;
	}

	public void disposeData() {
		lipsImage = null;
	}
}